import json
import datetime
import numpy as np
import pytz
import tzlocal
import pandas as pd
import logging
import collections

from cachetools import cached, LRUCache, TTLCache

@cached(cache=TTLCache(maxsize=1024, ttl=300))
def get_es_info(es):
    logger = logging.getLogger()
    logger.debug('get_es_info')
    return es.info()


class DateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()

        elif isinstance(o, datetime.time):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)


def elastic_to_dataframe(es, index, query="*", start=None, end=None, sort=None, timestampfield="@timestamp", datecolumns=[], _source=[], size=None, scrollsize=5000):
    """Convert an elastic collection to a dataframe.
    Parameters:
    es -- The elastic connection object
    query -- (optional) The elastic query
    start -- (optional) The time range start if any
    end -- (optional) The time range start if any
    sort -- (optional) The column we want to sort on
    timestampfield -- (optional) The timestamp field used by the start and stop parameters
    datecolumns -- (optional) A collection of columns that must be converted to dates
    _source -- (optional) columns we want to retrieve
    size -- (optional) The max number of recrods we want to retrieve
    scrollsize -- (optional) The size of the scroll to use
    """

    logger = logging.getLogger()
    array = []
    recs = []
    scroll_ids=[]


    version = int(get_es_info(es).get('version').get('number').split('.')[0])

    finalquery = {
        "_source": _source,
        "query": {
            "bool": {
                "must": [
                    {
                        "query_string": {
                            "query": query,
                            "analyze_wildcard": True
                        }
                    }
                ]
            }
        }
    }

    if start is not None:
        finalquery["query"]["bool"]["must"].append({
            "range": {

            }
        })

        finalquery["query"]["bool"]["must"][len(finalquery["query"]["bool"]["must"])-1]["range"][timestampfield] = {
            "gte": int(start.timestamp())*1000,
            "lte": int(end.timestamp())*1000,
            "format": "epoch_millis"
        }

    if sort is not None:
        finalquery["sort"] = sort

    logger.debug(finalquery)

    if size is not None and size < scrollsize:
        scrollsize = size

    res = es.search(index=index, size=scrollsize, scroll='1m', body=finalquery
                    )

    sid = res['_scroll_id']
    scroll_ids.append(sid)
    scroll_size = None
    if version < 7:
        scroll_size = res['hits']['total']
    else:
        scroll_size = res['hits']['total']['value']
        

    array = []
    for res2 in res["hits"]["hits"]:
        res2["_source"]["_id"] = res2["_id"]
        res2["_source"]["_index"] = res2["_index"]

        array.append(res2["_source"])

    recs = len(res['hits']['hits'])

    break_flag = False

    while (scroll_size > 0):
        res = es.scroll(scroll_id=sid, scroll='2m')
        sid = res['_scroll_id']
        scroll_ids.append(sid)
        scroll_size = len(res['hits']['hits'])
        logger.info("scroll size: " + str(scroll_size))
        logger.info("Next page:"+str(len(res['hits']['hits'])))
        recs += len(res['hits']['hits'])

        for res2 in res["hits"]["hits"]:
            if size is not None and len(array) >= size:
                break_flag = True
                break

            res2["_source"]["_id"] = res2["_id"]
            res2["_source"]["_index"] = res2["_index"]
            array.append(res2["_source"])

        if break_flag:
            break

    df = pd.DataFrame(array)

    if len(datecolumns) > 0 and len(df) > 0:
        containertimezone = pytz.timezone(tzlocal.get_localzone().zone)

        for col in datecolumns:
            if col not in df.columns:
                df[col] = None
            else:
                if df[col].dtype == "int64":
                    df[col] = pd.to_datetime(
                        df[col], unit='ms', utc=True).dt.tz_convert(containertimezone)
                else:
                    df[col] = pd.to_datetime(
                        df[col], utc=True).dt.tz_convert(containertimezone)
                    
    es.clear_scroll(body={'scroll_id': scroll_ids})
    return df


def dataframe_to_elastic(es, df, doc_type='doc'):
    """Converts a dataframe to an elastic collection to.
    The dataframe must have an "_index" column used to select the target index.
    Optionally an "_id" column can be used to specify the id of the record.
    Optionally an "_timestamp" column can be used to specify a "@timestamp column.

    Parameters:
    es -- The elastic connection object
    df -- The dataframe
    """

    logger = logging.getLogger(__name__)

    logger.debug("LOADING DATA FRAME")
    logger.debug("==================")

    version = int(get_es_info(es).get('version').get('number').split('.')[0])


    if len([item for item, count in collections.Counter(df.columns).items() if count > 1]) > 0:
        logger.error("NNOOOOOOOOBBBB DUPLICATE COLUMN FOUND "*10)
        raise Exception('Duplicate column in DataFrame')

    reserrors = []

    if len(df) == 0:
        logger.warning('dataframe empty')
    else:
        logger.debug("Loading data frame. Rows:" +
                    str(df.shape[1]) + " Cols:" + str(df.shape[0]))

    bulkbody = ""

    #df.columns=[_.replace("_timestamp","@timestamp") for _ in df.columns]
    df.rename(columns={"_timestamp": "@timestamp"},inplace=True)

    df_json = json.loads(df.to_json(orient='records'))
    
    action = {}
    action["index"] = {"_type": doc_type}

    flag_unique_index = False

    if len(df['_index'].unique() == 1):
        flag_unique_index = True
        
        action["index"]["_index"] = df.iloc[0]["_index"]
        
    for row in df_json:
        if "_id" in row:
            action["index"]["_id"] = row["_id"]
            
        if not flag_unique_index:
            action["index"]["_index"] = row["_index"]
            
        bulkbody += json.dumps(action, cls=DateTimeEncoder) + "\r\n"
        bulkbody += json.dumps({k:v for k,v in row.items() if k!='_id' and k!='_index' and v is not None}, 
                            cls=DateTimeEncoder) + "\r\n"


        if len(bulkbody) > 512000:
            logger.debug("BULK READY:" + str(len(bulkbody)))
            # print(bulkbody)
            bulkres = es.bulk(bulkbody, request_timeout=30)
            logger.debug("BULK DONE")
            bulkbody = ""

            if(not(bulkres["errors"])):
                logger.info("BULK done without errors.")
            else:
                for item in bulkres["items"]:
                    if "error" in item["index"]:
                        reserrors.append(
                            {"error": item["index"]["error"], "id": item["index"]["_id"]})

    if len(bulkbody) > 0:
        logger.debug("BULK READY FINAL:" + str(len(bulkbody)))
        bulkres = es.bulk(bulkbody)
        logger.debug("BULK DONE FINAL")

        if(not(bulkres["errors"])):
            logger.info("BULK done without errors.")
        else:
            for item in bulkres["items"]:
                if "error" in item["index"]:
                    reserrors.append(
                        {"error": item["index"]["error"], "id": item["index"]["_id"]})

    if len(reserrors) > 0:
        logger.warning(reserrors)



    return {
        'reserrors': reserrors
    }



import os
import time
import json
import time
import datetime
import requests
import unittest
import logging
import urllib3
import pandas as pd
import elasticsearch
from datetime import datetime
from datetime import timedelta
from elasticsearch import helpers
from elastic_helper import es_helper

from elasticsearch import Elasticsearch as ES, RequestsHttpConnection as RC

# git tag 1.0.1 -m "PyPi tag"
# git push --tags origin master
# python setup.py sdist
# twine upload dist/*

# python -m unittest  elastic_helper.es_helper_test

logger = None


server = os.environ["ES_HOST_TEST"]
user = os.environ["ES_USER_TEST"]
pwd = os.environ["ES_PWD_TEST"]


class TestClient(unittest.TestCase):
    """
    Test helper
    """

    def test_elastic_to_panda(self):
        """
        Send Receive
        """

        global server, pwd, user

        print("==> test_elastic_to_panda")
        try:
            host_params1 = {'host': server,
                            'port': 9200, 'use_ssl': True}  # RPINUM

            es = ES([host_params1], connection_class=RC, http_auth=(
                user, pwd), use_ssl=True, verify_certs=False)

            print(es.info())
            res2 = es_helper.elastic_to_dataframe(es, index="docker_stats*", scrollsize=1000, datecolumns=[
                                                  "read"], timestampfield="read", start=datetime.now()-timedelta(hours=1), end=datetime.now())
            res2["read"].describe()
            print(len(res2))

            self.assertTrue(res2 is not None)
            self.assertTrue(len(res2) > 0)
        finally:
            pass

    def test_size(self):

        global server, pwd, user

        print("==> test_size")
        try:
            host_params1 = {'host': server,
                            'port': 9200, 'use_ssl': True}  # RPINUM

            es = ES([host_params1], connection_class=RC, http_auth=(
                user, pwd), use_ssl=True, verify_certs=False)

            print(es.info())
            res2 = es_helper.elastic_to_dataframe(es, index="docker_stats*", size=10, timestampfield="read",
                                                  start=datetime.now()-timedelta(hours=1), end=datetime.now())

            print(len(res2))

            self.assertTrue(res2 is not None)
            self.assertTrue((len(res2) >= 0) and (len(res2) <= 10))
        finally:
            pass

    def test_size_2(self):

        global server, pwd, user

        print("==> test_size_2")
        try:
            host_params1 = {'host': server,
                            'port': 9200, 'use_ssl': True}  # RPINUM

            es = ES([host_params1], connection_class=RC, http_auth=(
                user, pwd), use_ssl=True, verify_certs=False)

            print(es.info())
            res2 = es_helper.elastic_to_dataframe(es, index="docker_stats*", size=345, scrollsize=100, timestampfield="read",
                                                  start=datetime.now()-timedelta(hours=10), end=datetime.now())

            print(len(res2))

            self.assertTrue(res2 is not None)
            self.assertTrue((len(res2) == 345))
        finally:
            pass

    def test_size_3(self):

        global server, pwd, user

        print("==> test_size_3")
        try:
            host_params1 = {'host': server,
                            'port': 9200, 'use_ssl': True}  # RPINUM

            es = ES([host_params1], connection_class=RC, http_auth=(
                user, pwd), use_ssl=True, verify_certs=False)

            print(es.info())
            res2 = es_helper.elastic_to_dataframe(es, index="docker_stats*", size=110, scrollsize=300, timestampfield="read",
                                                  start=datetime.now()-timedelta(hours=10), end=datetime.now())

            print(len(res2))

            self.assertTrue(res2 is not None)
            self.assertTrue((len(res2) == 110))
        finally:
            pass

    def test_date_cols(self):
        global server, pwd, user

        print("==> test_date_cols")

        try:
            host_params1 = {'host': server,
                            'port': 9200, 'use_ssl': True}  # RPINUM

            es = ES([host_params1], connection_class=RC, http_auth=(
                user, pwd), use_ssl=True, verify_certs=False)

            print(es.info())

            try:
                es.indices.delete('test_date_cols')
            except elasticsearch.NotFoundError:
                pass
            
            doc = {
                'attr1': 'test'
            }
            es.index(index="test_date_cols", id="t1", doc_type='_doc', body=doc)
            
            time.sleep(1)
            res = es_helper.elastic_to_dataframe(es, index="test_date_cols", datecolumns=["date1"])

            print(len(res))
            print(res.columns)

            self.assertTrue("date1" in res.columns)

            doc = {
                'attr1': 'test',
                'date1': datetime.now()
            }

            es.index(index="test_date_cols", id="t2", doc_type='_doc', body=doc)
            
            time.sleep(1)
            res = es_helper.elastic_to_dataframe(es, index="test_date_cols", datecolumns=["date1"])
            
            self.assertTrue("date1" in res.columns)

            doc = {
                'attr1': 'test',
                'date2': datetime.now()
            }

            es.index(index="test_date_cols", id="t3", doc_type='_doc', body=doc)
            
            time.sleep(1)
            res = es_helper.elastic_to_dataframe(es, index="test_date_cols", datecolumns=["date1", "date2"])
            
            self.assertTrue("date1" in res.columns)
            self.assertTrue("date2" in res.columns)

            es.indices.delete('test_date_cols')
        finally:
            pass
    
    def test_empty_attr(self):
        global server, pwd, user

        print("==> test_empty_attr")

        try:
            host_params1 = {'host': server,
                            'port': 9200, 'use_ssl': True}  # RPINUM

            es = ES([host_params1], connection_class=RC, http_auth=(
                user, pwd), use_ssl=True, verify_certs=False)

            print(es.info())

            try:
                es.indices.delete('test_empty_attr')
            except elasticsearch.NotFoundError:
                pass

            arr = [
                {
                    '_id': 't1',
                    'attr1': 'test'
                },
                {
                    '_id': 't2',
                    'attr2': 'test'
                },
                {
                    '_id': 't3',
                    'attr3': 'test'
                },
            ]

            df = pd.DataFrame(arr)
            
            INDEX_NAME = 'test_empty_attr'
            
            df['_index'] = INDEX_NAME
            
            es_helper.dataframe_to_elastic(es, df)
            time.sleep(1)

            rec = es.get(index = INDEX_NAME, id = 't1', doc_type = 'doc')

            self.assertTrue("attr1" in rec['_source'])
            self.assertTrue("attr2" not in rec['_source'])
            self.assertTrue("attr3" not in rec['_source'])
        finally:
            pass


from distutils.core import setup
setup(
  name = 'elastic_helper',
  packages = ['elastic_helper'], # this must be the same as the name above
  version = '1.2.2',
  description = 'Elastic Search to Pandas Dataframe',
  author = 'snuids',
  author_email = 'snuids@mannekentech.com',
  url = 'https://github.com/snuids/elastic_helper', 
  download_url = 'https://github.com/snuids/elastic_helper/archive/1.2.2.tar.gz',
  keywords = ['ElasticSearch', 'pandas', 'convert'], # arbitrary keywords
  classifiers = [],
)
